// Fermat/Double-Mersenne factoring program
//
// Version 2.4 - Mark Rodenkirch (rogue@wi.rr.com)
//
// Version History
//   1.0 - 01/28/2002 - Initial Release
//   1.1 - 02/04/2002 - Added checkpointing
//   1.2 - 07/12/2003 - New sieve (based upon primegen) and change to
//                      main loop to improve speed up to 80%.
//   1.3 - 07/20/2003 - Finish checkpointing.  Tested on x86 (K6, K7, and P-III) and PPC (G4)
//   1.4 - 07/24/2003 - Use rounding in mulmod based upon CPU
//   1.5 - 08/05/2003 - Create output file containing found factors
//   1.6 - 06/14/2004 - Fix minor bug in allocateMemory where it checked wrong pointer after malloc
//   1.7 - 09/22/2005 - Fix all compiler bugs and fixed a bug that occurs when kStart > MASK_SIZE
//                      that causes the sieve to not work correctly.
//   2.0 - 12/29/2005 - Rewrote the sieve and limit k and n to under 64 bits.
//   2.1 - 04/04/2008 - Added support for Double-Mersenne checks
//   2.2 - 12/08/2008 - Fixed in bug in Evaluate() that occurs when GMP is 32-bit
//                      and k > 2^32.
//   2.3 - 12/28/2010 - Only check Mersenne primes for Double-Mersenne checks.  Fix output for
//                      found factors.  Add more error checking.  Fixed computation of progress.
//                      Checkpoint every 60 seconds.  Fix output for WIN32 builds as WIN32 uses
//                      different ASCII escape sequences.  Fix output when checkpointing.  Fix
//                      bug where kEnd wasn't evaluated.
//   2.4 - 01/01/2011 - Fixed bug where only first 1e6 of range would be tested.
//   2.5 - 07/16/2016 - Fixed a bug in asm routines for 768 bin values.

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <gmp.h>

#define FERMAT_VERSION         "2.5"

// This is the number of primes to sieve along with those primes.  The amount
// of memory allocated during execution is approximately 16x this value.
uint32_t  primeCount;
uint32_t *primeTable;

// This table contains the smallest k where k*2^n+1 = 0 mod p.  All values
// of k for k (k+p), (k+2p), ... are composite and will not be evaluated
uint32_t *residualTable;

// This table is the same size of primeTable.  The values in this table
// correspond to the smallest composite greater than the low end of the
// range being sieved.
uint64_t *compositeTable;

// RANGE_SIZE contains the current range of candidates that has been
// sieved.  The size of this variable should not exceed the size
// of level 2 cache as that will dramatically hurt performance.
#define RANGE_SIZE             500000

uint64_t  kStart, kCurrent, kEnd, kTested, startingK;
int32_t   nStart, nCurrent, nEnd;
mpz_t     temp, fermat, nTemp, kTemp, factor, minus1;
double    startTime, endTime;
double    pprime, preciprocal;

/* function prototypes */
char *ReadIni();
void  GetResiduals(uint32_t n);
void  InitializePrimeSieve();
void  Sieve(uint64_t lowK, uint64_t highK);
void  Evaluate(uint64_t kToTest);
void  CheckPoint(uint64_t currentK);

#if defined(WIN32)
   // 3 * 2^62
   const double rounding = 3.0 * (double) (1 << 20) * (double) (1 << 21) * (double) (1 << 21);
   #define PRINTF_U64   "%I64u"
#elif defined(__ppc__) || defined(__ppc64__)
   // 2^52
   const double rounding = (double) (1 << 26) * (double) (1 << 26);
   #define PRINTF_U64   "%llu"
#endif

// (a * b) | pprime
__inline double mulmod(double a, double b)
{
    double c, d, result;

    c = d = a * b * preciprocal;
    // Round c to nearest integer
    c += rounding;
    c -= rounding;

    // make sure d is larger so we can get the remainder
    if (c > d)
       d += 1.0;

    result = (d - c) * pprime;
    result += rounding;
    result -= rounding;

    return result;
}

// (x ^ e) | pprime
__inline double expmod(double x, int e)
{
    double y = 1.0;

    for (;;)
    {
        if (e & 1)
            y = mulmod(x, y);

        e >>= 1;
        if (!e)
            return y;

        x = mulmod(x, x);
    }
}

#ifdef _DM_
int32_t mplist[] = { 2, 3, 5, 7, 13, 17, 19, 31, 61, 89, 107, 127, 521, 607, 1279, 2203,
                     2281, 3217, 4253, 4423, 9689, 9941, 11213, 19937, 21701, 23209, 44497,
                     86243, 110503, 132049, 216091, 756839, 859433, 1257787, 1398269, 2976221,
                     3021377, 6972593, 13466917, 20996011, 24036583, 25964951, 30402457, 32582657,
                     37156667, 42643801, 43112609, 57885161, 74207281, 0 };
#endif

int main(int argc, char *argv[])
{
   uint64_t possibleK;
   double   milliSeconds, testedPct, testedRate;
   char    *msg;

#ifdef _DM_
   printf("GMP-Double-Mersenne version %s\n\n", FERMAT_VERSION);
#else
   printf("GMP-Fermat version %s\n\n", FERMAT_VERSION);
#endif
   msg = ReadIni();
   if (msg) 
   {
      printf("%s\nExiting\n", msg);
      exit(0);
   }
   InitializePrimeSieve();

   mpz_init(temp);
   mpz_init(fermat);
   mpz_init(nTemp);
   mpz_init(kTemp);
   mpz_init(factor);
   mpz_init(minus1);

   while (nCurrent <= nEnd)
   {
#ifdef _DM_
      int32_t *mp;

      mp = mplist;
      while (*mp && *mp < nCurrent)
         mp++;

      if (nCurrent != *mp)
      {
         nCurrent++;
         continue;
      }
#endif

      printf("\nTesting for n = %d\n", nCurrent);

      mpz_set_ui(nTemp, 2);
      mpz_pow_ui(nTemp, nTemp, nCurrent);
      mpz_set_ui(fermat, 2);

      if (GMP_LIMB_BITS == 64)
         mpz_set_ui(kTemp, kCurrent);
      else
      {
         mpz_set_ui(kTemp, kCurrent >> 32);
         mpz_mul_2exp(kTemp, kTemp, 32);
         mpz_add_ui(kTemp, kTemp, kCurrent & (0xffffffff));
      }

      GetResiduals(nCurrent);
      kTested = 0;
      startingK = kCurrent;
      startTime = (double) clock() / CLOCKS_PER_SEC;

      Sieve(kCurrent, kEnd);

      endTime = ((double) clock() / CLOCKS_PER_SEC);
      milliSeconds = endTime - startTime;
      possibleK = 1 + (kEnd - startingK);
      testedPct = (100.0 * (double) kTested) / (double) possibleK;
      testedRate = (double) possibleK / milliSeconds;

      printf("\nOverall rate for n=%d:  tested "PRINTF_U64" of "PRINTF_U64" numbers (%5.2f pct) in %8.2f seconds at %.0lf k per second\n",
              nCurrent, kTested, possibleK, testedPct, milliSeconds, testedRate);

      kCurrent = kStart;
      nCurrent++;
   }

   printf("\nDone\n");

   free(primeTable);
   free(compositeTable);
   free(residualTable);

   CheckPoint(0);
   return 0;
}

// Find the value k where k*2^n+1 = 0 mod p.  This will allow us to sieve 
// out k and k+x*p for all x as those values are not prime.
//
// The fastest algorithm for finding value k0 , that satisfies the condition k0. 
// 2n+1=0 (mod p), has been recently offered by Paul Jobling. He has written:
// There is an approach that is independent of the size of p and has a speed that 
// is O(log n), so the speed is fixed regardless of the size of p
//
// p | k*(2^n-1)+1 --> k*(2^n-1)+1 = 0 (mod p)
//                 --> k*x = p-1 (mod p)
//
// p | k*2^n+1 -->  k*2^n+1 = 0 (mod p)
//             -->  k = -1/(2^n)(mod p)
//             -->  -k = (1/2)^n (mod p)
//
//                  1/2 (mod p) = (p+1)/2 (mod p)
//             -->  -k = ((p+1)/2)^n (mod p)
//
// So just do a left-to-right exponentiation (mod p) of (p+1)/2 to the power n.
// 
// For example, if n=13 and p=17: 
// (p+1)/2 = 9, and 9^13 (mod 17) = 8.  So if k = -8(mod 17), then 
// 17 | k*2^13+1. And indeed 17 does divide (-8+17).2^13+1, (9+17).2^13+1, etc.

void   GetResiduals(uint32_t nCurrent)
{ 
   uint32_t  i, r, x, n;

#ifdef _DM_
   // Double-Mersenne code here
   for (i=0; i<primeCount; i++)
   {
      pprime = primeTable[i];
      preciprocal = 1 / (double) primeTable[i];

      x = expmod(2, nCurrent);
      if (x != 1)
      {
         if (x == 0)
         {
            x = primeTable[i] - 1;
            r = 1;
         }
         else
         {
            x--;
            n = x;
            r = 1;
            while (n != primeTable[i]-1)
            {
               r++;
               if (n+x >= primeTable[i])
                  n = n + x - primeTable[i];
               else
                  n += x;

            }
         }

         residualTable[i] = r;
      }
      else
         residualTable[i] = 0;

#ifdef DEBUG
      printf("(%d*(2^%d-1)+1)%%%d\n", r, nCurrent, primeTable[i]);
#endif
   } 
         
#else
   // Fermat code here
   for (i=0; i<primeCount; i++)
   {
      x = (primeTable[i]+1) >> 1;
      n = nCurrent % (primeTable[i]-1);
      if (n == 0)
         r = 1;
      else
      {
         pprime = primeTable[i];
         preciprocal = 1 / (double) primeTable[i];
         r = (uint32_t) expmod(x, n);
      }

      residualTable[i] = primeTable[i] - r;

#ifdef DEBUG
      printf("(%d*2^%d+1)%%%d\n", residualTable[i], nCurrent, primeTable[i]);
#endif
   }
#endif
}

char  *ReadIni()
{
   FILE    *ini;
   char     input[100];

   kStart = kCurrent = kEnd = 0;
   nStart = nCurrent = nEnd = 0;
   primeCount = 0;

   ini = fopen("fermat.ini", "r");
   if (ini == 0)
   {
      printf("The file fermat.ini could not be opened or was not found.\nExiting program.\n");
      exit(0);
   }

   while (fgets(input, sizeof(input), ini) > 0)
   {
      if (!memcmp(input, "kStart=", 7))
         kStart = atoll(&input[7]);
      if (!memcmp(input, "kCurrent=", 9))
         kCurrent = atoll(&input[9]);
      if (!memcmp(input, "kEnd=", 5))
         kEnd = atoll(&input[5]);
      if (!memcmp(input, "nStart=", 7))
         nStart = atol(&input[7]);
      if (!memcmp(input, "nCurrent=", 9))
         nCurrent = atol(&input[9]);
      if (!memcmp(input, "nEnd=", 5))
         nEnd = atol(&input[5]);
      if (!memcmp(input, "FilterPrimes=", 13))
         primeCount = atol(&input[13]);
   }
   fclose(ini);

   if (kStart == 0)
      return "The starting value for k (kStart) was not set in fermat.ini.";

   if (kEnd == 0)
      return "The starting value for k (kEnd) was not set in fermat.ini.";

   if (kStart > kEnd)
      return "The ending value for k (kEnd) is less than the starting value of k (kStart).";

   if ((kCurrent > 0) && (kCurrent < kStart || kCurrent > kEnd))
      return "The current value for k (kCurrent) is outside of the range of k.";

   if (nStart == 0)
      return "The starting value for n (nStart) was not set in fermat.ini.";

   if (nEnd == 0)
      return "The starting value for n (nEnd) was not set in fermat.ini.";

   if (nStart > nEnd)
      return "The ending value for n (nEnd) is less than the starting value of n (nStart).";

   if ((nCurrent > 0) && (nCurrent < nStart || nCurrent > nEnd))
      return "The current value for n (nCurrent) is outside of the range of n.";

#ifdef _DM_
   if (kStart & 1)
   {
      if (kStart == kEnd)
         return "For Double-Mersenne factoring, kStart from fermat.ini must be even.  Since kEnd has the same value, there is nothing to test.";

      printf("kStart from fermat.ini is odd, it will be made even.\n");
      kStart++;
   }
#else
   if (!(kStart & 1))
   {
      if (kStart == kEnd)
         return "For Fermat factoring, kStart from fermat.ini must be odd.  Since kEnd has the same value, there is nothing to test.";

      printf("kStart from fermat.ini is even, it will be made odd.\n");
      kStart++;
   }
#endif

   if (kCurrent == 0)
      kCurrent = kStart;
#ifdef _DM_
   else if (kCurrent & 1)
   {
      printf("kCurrent from fermat.ini is odd, it will be made even.\n");
      kCurrent++;
   }
#else
   else if (!(kCurrent & 1))
   {
      printf("kCurrent from fermat.ini is even, it will be made odd.\n");
      kCurrent++;
   }
#endif

   if (primeCount == 0)
   {
      printf("The value FilterPrimes was not set, it will default to 500000.\n");
      primeCount = 500000;
   }

   if (nCurrent == 0)
      nCurrent = nStart;

   if ((kCurrent != kStart) || (nCurrent != nStart))
      printf("Restarting range of k from "PRINTF_U64" to "PRINTF_U64" at "PRINTF_U64" and range of n from %d to %d at %d\n",
              kStart, kEnd, kCurrent, nStart, nEnd, nCurrent);
   else
      printf("Processing range of k from "PRINTF_U64" to "PRINTF_U64" and range of n from %d to %d\n",
              kStart, kEnd, nStart, nEnd);

   return 0;
}

void    InitializePrimeSieve()
{
   uint32_t  sqrtMax, i, composite, maxComposite, primesInPrimeTable;
   uint32_t  p, minp, *lowPrimes, lowPrimeCount;
   uint8_t  *sieve, *sievePtr;

   // Assume that p(primeCount) < 30 * primeCount
   maxComposite = 30 * primeCount;

   sqrtMax = (uint32_t) sqrt((double) maxComposite) + 10;

   // Make sure this is large enough to hold all of the low
   // primes we need.
   lowPrimes = (uint32_t *) malloc(1000000 * sizeof(uint32_t));
   lowPrimes[0] = 3;
   lowPrimeCount = 1;
   for (p=5; p<sqrtMax; p+=2)
   {
      for (minp=0; minp<=lowPrimeCount; minp++)
      {
         if (lowPrimes[minp] * lowPrimes[minp] > p)
         {
            lowPrimes[lowPrimeCount] = p;
            lowPrimeCount++;
            break;
         }
         if (p % lowPrimes[minp] == 0)
             break;
      }
   }

   // Divide maxComposite by 2 to save memory, also because already know
   // that all even numbers in the sieve are composite
   sieve = (uint8_t *) malloc((maxComposite >> 1) * sizeof(uint8_t));
   if (!sieve)
   {
      printf("There is not enough memory to allocate a table for the sieve.  Program terminated\n");
      exit(0);
   }

   memset(sieve, 1, (maxComposite >> 1));

   for (i=0; i<lowPrimeCount; i++)
   {
      // Get the current low prime.  Start sieving at 3x that prime
      // since 1x is prime and 2x is divisible by 2.
      // sieve[1] = 3, sieve[2] = 5, etc.
      composite = lowPrimes[i] * 3;
      sievePtr = &sieve[(composite - 1) >> 1];

      while (composite < maxComposite)
      { 
         // composite will always be odd, so add 2*lowPrimes[i]
         *sievePtr = 0;
         sievePtr += lowPrimes[i];
         composite += (lowPrimes[i] << 1);
      } 
   }

   primeTable = (uint32_t *) malloc(primeCount * sizeof(uint32_t));
   residualTable = (uint32_t *) malloc(primeCount * sizeof(uint32_t));

   if (!primeTable || !residualTable)
   {
      printf("There is not enough memory to allocate table of primes.  Program terminated\n");
      exit(0);
   }

   primesInPrimeTable = 0;
   for (i=1; i<(maxComposite >> 1); i++)
   {
      if (sieve[i])
      {
         // Convert the value back to an actual prime
         primeTable[primesInPrimeTable] = (i << 1) + 1;
         primesInPrimeTable++;
         if (primesInPrimeTable == primeCount)
            break;
      }
   }

   free(sieve);
   free(lowPrimes);
}

void    Sieve(uint64_t lowK, uint64_t highK)
{
   uint64_t  composite, prime, currentK, candidate, lastComposite, maxPrime;
   uint32_t  i;
   uint8_t  *sieve, *sievePtr;
   time_t    lastCheckPointTime;

   compositeTable = (uint64_t *) malloc((primeCount + 1) * sizeof(uint64_t));
   sieve = (uint8_t *) malloc((RANGE_SIZE + 1) * sizeof(uint8_t));
   currentK = lowK;

   if (!compositeTable || !sieve)
   {
      printf("There is not enough memory to allocate the composite table.  Program terminated\n");
      exit(0);
   }

   // Find the largest composite greater than lowK
   for (i=0; i<primeCount; i++)
   {
      maxPrime = primeTable[i];
      lastComposite = (lowK / maxPrime) * maxPrime;
      compositeTable[i] = lastComposite + maxPrime + residualTable[i];

#ifdef _DM_
      // We only care about even composites since the 
      // sieve range only refers to even values
      if (compositeTable[i] & 1)
#else
      // We only care about odd composites since the 
      // sieve range only refers to odd values
      if (!(compositeTable[i] & 1))
#endif
         compositeTable[i] += maxPrime;

#ifdef DEBUG
      if (compositeTable[i] < highK)
         printf("(%lld*(2^%d-1)+1)%%%d\n", compositeTable[i], nCurrent, primeTable[i]);
#endif
   }

   lastCheckPointTime = time(NULL);
   while (currentK <= highK)
   {
      memset(sieve, 1, RANGE_SIZE);

      for (i=0; i<primeCount; i++)
      {
         prime = primeTable[i];
         composite = compositeTable[i];
         sievePtr = &sieve[(composite - currentK) >> 1];

         while (composite < currentK + (RANGE_SIZE << 1))
         {
            *sievePtr = 0;
            sievePtr += prime;
            composite += (prime << 1);
         }

         compositeTable[i] = composite;
     }

     candidate = currentK;

     for (i=0; i<RANGE_SIZE; i++, candidate+=2)
     {
       if (candidate > highK)
         break;

       if (candidate >= lowK && sieve[i])
       {
          Evaluate(candidate);

          if (time(NULL) > lastCheckPointTime + 60)
          {
             CheckPoint(candidate);
             lastCheckPointTime = time(NULL);
          }
       }
    }

    currentK += (RANGE_SIZE << 1);
  }
}

void  Evaluate(uint64_t kToTest)
{
   FILE   *found;
#ifndef _DM_
   int32_t m;
#endif

   kTested++;

   if (GMP_LIMB_BITS == 64)
     mpz_set_ui(kTemp, kToTest);
   else
   {
     mpz_set_ui(kTemp, kToTest >> 32);
     mpz_mul_2exp(kTemp, kTemp, 32);
     mpz_add_ui(kTemp, kTemp, kToTest & (0xffffffff));
   }
   mpz_mul_2exp(minus1, kTemp, nCurrent);
   mpz_add_ui(factor, minus1, 1);

#ifdef _DM_
   mpz_sub(factor, factor, kTemp);
#endif

   mpz_powm(temp, fermat, nTemp, factor);

#ifdef _DM_
   if (mpz_cmp_ui(temp, 2) == 0)
   {
#ifdef WIN32
      printf("\nFound factor "PRINTF_U64"*(2^%d-1)+1 of 2^(2^%d-1)-1\n", kToTest, nCurrent, nCurrent);
#else
      printf("\n\033[7mFound factor "PRINTF_U64"*(2^%d-1)+1 of 2^(2^%d-1)-1\033[0m\n", kToTest, nCurrent, nCurrent);
#endif
      found = fopen("factors.txt", "a+");
      fprintf(found, "Found factor "PRINTF_U64"*(2^%d-1)+1 of 2^(2^%d-1)-1\n", kToTest, nCurrent, nCurrent);
      fclose(found);
   }
#else
   if (mpz_cmp(temp, minus1) == 0)
   {
#ifdef WIN32
      printf("\nFound factor "PRINTF_U64"*2^%d+1 of 2^(2^%d)+1\n", kToTest, nCurrent, nCurrent-2);
#else
      found = fopen("factors.txt", "a+");
#endif
      fprintf(found, "Found factor "PRINTF_U64"*2^%d+1 of 2^(2^%d)+1\n", kToTest, nCurrent, nCurrent-2);
      fclose(found);
   }
   else if (mpz_cmp_ui(temp, 1) == 0)
   {
      mpz_set_ui(temp, 2);
      for (m=1; m<=nCurrent-2; m++)
      {
         mpz_powm_ui(temp, temp, 2, factor);
         if (mpz_cmp(temp, minus1) == 0)
         {
#ifdef WIN32
            printf("\nFound factor "PRINTF_U64"*2^%d+1 of 2^(2^%d)+1\n", kToTest, nCurrent, m);
#else
            printf("\n\033[7mFound factor "PRINTF_U64"*2^%d+1 of 2^(2^%d)+1\033[0m\n", kToTest, nCurrent, m);
#endif
            found = fopen("factors.txt", "a+");
            fprintf(found, "Found factor "PRINTF_U64"*2^%d+1 of 2^(2^%d)+1\n", kToTest, nCurrent, m);
            fclose(found);
         }
      }
   }
#endif

   kCurrent = kToTest;
}

void  CheckPoint(uint64_t currentK)
{
   FILE    *ini;
   uint64_t possibleK;
   double   milliSeconds, testedPct, testedRate;

   ini = fopen("fermat.ini", "w");

   if (ini == 0)
   {
      printf("The file fermat.ini could not be opened or was not found.\nExiting program.\n");
      exit(0);
   }

   fprintf(ini, "kStart="PRINTF_U64"\n", kStart);
   if (currentK != 0)
      fprintf(ini, "kCurrent="PRINTF_U64"\n", kCurrent);
   fprintf(ini, "kEnd="PRINTF_U64"\n", kEnd);
   fprintf(ini, "nStart=%d\n", nStart);
   if (currentK != 0)
      fprintf(ini, "nCurrent=%d\n", nCurrent);
   fprintf(ini, "nEnd=%d\n", nEnd);
   fprintf(ini, "FilterPrimes=%d\n", primeCount);

   fclose(ini);

   if (currentK != 0)
   {
      endTime = (double) clock() / CLOCKS_PER_SEC;
      milliSeconds = ((double) (endTime - startTime));
      possibleK = currentK - startingK;
      testedPct = (100.0 * (double) kTested) / (double) possibleK;
      testedRate = (double) possibleK / milliSeconds;

      printf("\rTested to "PRINTF_U64" (tested %5.2f pct candidates) at %.0lf per second   ", 
                currentK, testedPct, testedRate); 
      fflush(stdout);
   }
}
